/*
  yacc file for PCD format

  Written and Revised by Akihiro Tsutsui (NTT Optical Network Systems Labs.)
  Copyright (c) 1994, 1995 Nippon Telegraph & Telephone Corporation
*/

%{

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alloca.h>

#include "ulist.h"
#include "pcddefs.h"
#include "pcdparse.h"

#define YYDEBUG 1

extern char *yytext ;
extern int  yylex() ;
extern void yyerror(char *s) ;

/* string buffers */

static char dname_buf[64] ;
static char pin_name_buf[64] ;

static char* str_buf ;

/* PCD Variables */

extern PCD_MODULE_LIST* pcd_modules ;
extern int pcd_mod_count ;

static PCD_MODULE* pcd_module_p ;

static PCDPinList* pcd_pin_list ;
static PCDCmdList* pcd_spc_list ;

static PCDPinInfo  pcd_pinfo_buf ;
static PCDPinInfo* pcd_pinfo_p ;

static PCDStrList* pcd_slist_p ;

%}

%union {
	char *strval ;
	float *fval ;
	int  ival ;
	PCDPinType pin_type ;
}

%type <strval> identifier
%type <strval> s_pin
%type <strval> f_pin
%type <strval> operator
%type <strval> dname
%type <fval>   value
%type <pin_type> ptype
%type <pin_type> pin_type

%token LPAREN RPAREN

%token DEF_MODULE
%token DEF_PIN
%token DEF_FUNCTION
%token DEF_LOGIC
%token DEF_DELAY
%token DEF_CONSTRAINT
%token DEF_SYMBOL
%token DEF_SPECIAL

%token POWER
%token AREA
%token GATES
%token CELL_NAME

%token LOAD
%token TYPE
%token NOTE

%token INPUT
%token OUTPUT
%token BIDIRECT
%token TRISTATE
%token UNUSED

%token F_NAME
/*
%token FREG_OP
%token AND_OP
%token OR_OP
*/
%token IDENTIFIER
%token FLOAT
%token INTEGER

%token NOT

%start pcd

%%
pcd				:	module_defs
				;

module_defs		:	/* empty */
				|	module_defs	module_def
				;

module_def		:	LPAREN DEF_MODULE dname
					{
	/* Initialize */
						pcd_pinfo_buf.clear() ;

	/* Making pcd module entry */

						pcd_module_p = new PCD_MODULE(pcd_mod_count++, $3) ;

						pcd_pin_list = new PCDPinList ;
						pcd_spc_list = new PCDCmdList ;

						pcd_module_p->pinlist = pcd_pin_list ;
						pcd_module_p->defspec = pcd_spc_list ;

						pcd_modules->append(pcd_module_p) ;

	/* If you want to use another information, make corresponding entry. */

					}
					mod_attrs defs RPAREN
				;

dname			:	identifier
					{
						$$ = dname_buf ;
						strcpy(dname_buf, $1) ;
					}
				;

mod_attrs		:	mod_attr
				|	mod_attrs mod_attr
				;

mod_attr		:	POWER p_value
				|	AREA p_value
				|	GATES p_value
				|	CELL_NAME cell_name
				;

cell_name		:	identifier
					{ }
				;

defs			:	/* empty */
				|	defs def
				;

def				:	pin_def
				|	function_def
				|	delay_def
				|	constraint_def
				|	symbol_def
				|	spc_def
				|	logic_def
				;

pin_def			:	LPAREN DEF_PIN pin_name pin_atr RPAREN
					{
						pcd_pinfo_p = new PCDPinInfo(pcd_pinfo_buf) ;
						pcd_pinfo_buf.clear() ;
						pcd_pin_list->append(pcd_pinfo_p) ;
					}
				;

pin_name		:	identifier
					{
						strcpy(pin_name_buf, $1) ;
						pcd_pinfo_buf.name = pin_name_buf ;
					}
				;

pin_atr			:	pin_type pin_load pin_note
					{
						pcd_pinfo_buf.type = $1 ;
					}
				|	pin_type pin_load
					{
						pcd_pinfo_buf.type = $1 ;
					}
				|	pin_type
					{
						pcd_pinfo_buf.type = $1 ;
					}
				;

pin_load		:	LOAD p_value
					{ }
				;

pin_type		:	TYPE ptype
					{ $$ = $2 ;}
				;

pin_note		:	NOTE identifier
					{ }
				;

ptype			:	INPUT
					{ $$ = PIN_INPUT ; }
				|	OUTPUT
					{ $$ = PIN_OUTPUT ; }
				|	BIDIRECT
					{ $$ = PIN_BIDIRECT ; }
				|	TRISTATE
					{ $$ = PIN_TRISTATE ; }
				|	UNUSED
					{ $$ = PIN_UNUSED ; }
				;

function_def	:	LPAREN DEF_FUNCTION f_pin f_form RPAREN
				;

f_forms			:	/* empty */
				|	f_forms f_form
				;

f_form			:	LPAREN operator f_forms RPAREN
				|	f_pin
					{ }
				;

f_pin			:	identifier
					{ }
				|	NOT identifier
					{ }
				;

operator		:	identifier
					{ }
				;

delay_def		:	LPAREN DEF_DELAY d_start d_end d_form RPAREN
				;

d_start			:	event
				;

d_end			:	event
				;

d_form			:	LPAREN '+' d_form1s RPAREN
				|	LPAREN '+' RPAREN
				|	d_form1
				;

d_form1s		:	d_form1
				|	d_form1s d_form1
				;

d_form1			:	LPAREN '*' value d_prm RPAREN
				|	LPAREN '*' value RPAREN
				|	d_prm
				;

d_prm			:	dpin_load
				|	value
					{}
				;

dpin_load		:	'~' pin_simple
				;

constraint_def	:	LPAREN DEF_CONSTRAINT identifier c_form c_type RPAREN
				;

c_form			:	LPAREN '-' c_form1s RPAREN
				|	LPAREN '-' RPAREN
				|	c_form1
				;

c_form1s		:	c_form1
				|	c_form1s c_form1
				;

c_form1			:	LPAREN '+' c_form2s RPAREN
				|	c_form2
				;

c_form2s		:	c_form2
				|	c_form2s c_form2
				;

c_form2			:	LPAREN '*' value c_prm RPAREN
				|	LPAREN '*' value RPAREN
				|	c_prm
				;

c_prm			:	dpin_load
				|	event
				|	variable
				|	value
					{}
				;

c_type			:	/* empty */
				|	TYPE identifier
				;

symbol_def		:	LPAREN DEF_SYMBOL symbol_name s_pins RPAREN
				;

symbol_name		:	identifier
					{ }
				;

s_pins			:	/* empty */
				|	s_pins s_pin
				;

s_pin			:	identifier
					{ }
				|	NOT identifier
					{ }
				;

spc_def			:	LPAREN DEF_SPECIAL
					{
						pcd_slist_p = new PCDStrList ;
						pcd_spc_list->append(pcd_slist_p) ;
					}
					sp_atoms RPAREN
				;

sp_atoms		:	/* empty */
				|	sp_atoms sp_atom
				;

sp_atom			:	identifier
					{
						str_buf = new char[strlen($1)+1] ;
						strcpy(str_buf, $1) ;
						pcd_slist_p->append(str_buf) ;
					}
				;

logic_def		:	LPAREN DEF_LOGIC logic_out logic_exp RPAREN
				;

logic_out		:	logic_out_pin
				|	NOT logic_out_pin
				;

logic_out_pin	:	identifier
					{ }
				;

logic_exps		:	logic_exp
				|	logic_exps logic_exp
				;

logic_exp		:	logic_atom
				|	NOT logic_atom
				|	LPAREN logic_op logic_exps RPAREN
				|	NOT LPAREN logic_op logic_exps RPAREN
				;

logic_op		:	identifier
					{ }
				;

logic_atom		:	'0'
				|	'1'
				|	'X'
				|	identifier
					{}
				;

event			:	'/' pin_simple
				|	'\\' pin_simple
				|	'%' pin_simple
				;

pin_simple		:	identifier
					{ }
				;

variable		:	'?' identifier
				;

identifier		:	IDENTIFIER
					{ $$ = yytext ; }
				;

p_value			:	value
					{} 
				;

value			:	FLOAT
					{}
				|	INTEGER
					{}
				;
%%

PCD_MODULE_LIST* pcd_modules ;
int pcd_mod_count ;

int 
parse_pcdx( FILE *fp, PCD_MODULE_LIST& PML )
{

	extern PCD_MODULE_LIST* pcd_modules ;
	extern int pcd_mod_count ;

	extern int yyparse() ;
	extern void yyrestart(FILE* f) ;
	extern FILE* yyin ;

	extern int pcdparse_lineno ;

	// initialize

	int retval ;
	static entflg = 0 ;
	pcd_mod_count = 0 ;
	pcd_modules = &PML ;
	
	pcdparse_lineno = 0 ;

	if (entflg) {
		yyrestart(fp) ;
		retval = yyparse() ;
	}
	else {
		yyin = fp ;
		retval = yyparse() ;
		entflg = 1 ;
	}

	if (retval) {
		yyerror("parse error") ;
		return (-1) ;
	}

	return pcd_mod_count ;
}
